#!/usr/bin/env python
# -*- coding:utf-8 -*-

import inspect
import pathlib as pt
import pprint

import cerberus
import jinja2
import yaml
from jinja2 import meta

from ioproc.exceptions import MissingOverridesError
from ioproc.logger import mainlogger as log
from ioproc.schemas import (action_schema, checkpoint_schema,
                            executable_schema, general_schema)

__author__ = ["Benjamin Fuchs", "Judith Vesper", "Felix Nitsch", "Jan Buschmann"]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = ["Niklas Wulff", "Hedda Gardian", "Gabriel Pivaro", "Kai von Krbek"]

__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


class ConfigurationError(Exception):
    """
    Error raised by the configuration process.
    """

    pass


class ConfigDict(dict):
    def __getitem__(self, fieldname):
        """
        Access the config dictionary and retrieve data set by field name.
        If an KeyError is raised and the actionName is None, a ConfigurationError
        is raised, warning the user of an unsuccessful configuration.
        :param field name
        :return data from config dictionary which is accessed using the field name
        """
        try:
            return super().__getitem__(fieldname)
        except KeyError:
            s = inspect.stack()

            actionName = None
            lineno = s[0].lineno
            filename = pt.Path(s[0].filename).name

            for idx, iframe in enumerate(s):
                if iframe.function == "__actionwrapper__":
                    last = s[idx - 1]
                    actionName = last.function
                    lineno = last.lineno
                    filename = pt.Path(last.filename).name
                    break

            if actionName is None:
                out = '\n      config field "{}" unavailable\n' '      in file "{}" in line {}'.format(
                    fieldname, filename, lineno
                )
            else:
                out = (
                    '\n      config field "{}" unavailable\n'
                    '      requested by action "{}" in line {}\n'
                    '      in file "{}"'.format(fieldname, actionName, lineno, filename)
                )
            raise ConfigurationError(out)

    def print(self):
        pprint.pprint(self)


class ConfigList(list):
    def __getitem__(self, fieldname):
        """
        Access the config list and retrieve data set by field name.
        If an KeyError is raised and the actionName is None, a ConfigurationError
        is raised, warning the user of an unsuccessful configuration.
        :param field name
        :return data from config list which is accessed using the field name
        """
        try:
            return super().__getitem__(fieldname)
        except (TypeError, IndexError):
            s = inspect.stack()

            actionName = None
            lineno = s[0].lineno
            filename = pt.Path(s[0].filename).name

            for idx, iframe in enumerate(s):
                if iframe.function == "__actionwrapper__":
                    last = s[idx - 1]
                    actionName = last.function
                    lineno = last.lineno
                    filename = pt.Path(last.filename).name
                    break

            if actionName is None:
                out = "\n      element at position {} unavailable\n" '      in file "{}" in line {}'.format(
                    fieldname, filename, lineno
                )
            else:
                out = (
                    "\n      element at position {} unavailable\n"
                    '      requested by action "{}" in line {}\n'
                    '      in file "{}"'.format(fieldname, actionName, lineno, filename)
                )

            raise ConfigurationError(out)


def convertList(d):
    """
    When loading the .yaml, the user defined workflow is stored in a dictionary.
    This method converts the dictionary to a list.
    :return list with actions to execute in workflow
    """
    if not isinstance(d, ConfigDict):
        for ikey, ivalue in d.items():
            if hasattr(ivalue, "keys"):
                d[ikey] = convertList(ivalue)
            elif not hasattr(ivalue, "strip") and hasattr(ivalue, "__iter__"):
                d[ikey] = ConfigList(ivalue)
    return d


def loadAndValidate(confPath, overridedata={}):
    """
    Loads the config file from a path provided and validates it against a provided schema. If the config file is
    empty, an empty dictionary is returned in order to comply with the following interfaces.
    :param confPath: path to config, schema: validation schema which ensure certain standards of configuration
    :return config dictionary which is validated against a provided schema
    """

    content = pt.Path(confPath).read_text()

    content = _correct_deprecated_fields(content, log)

    env = jinja2.Environment()
    ast = env.parse(content)
    undeclared_overrides = meta.find_undeclared_variables(ast)
    missing_overrides = undeclared_overrides.difference(set(overridedata.keys()))
    if len(missing_overrides) > 0:
        raise MissingOverridesError(f"Please provide the following missing overrides: {missing_overrides}")

    t = jinja2.Template(content)

    conf = t.render(**overridedata)
    conf = yaml.load(conf, Loader=yaml.Loader)

    if conf is None:
        return {}

    dirs = {}
    if "directives" in conf:
        dirs = conf["directives"]
        del conf["directives"]

    validate_config(conf, confPath)
    conf = convertList(conf)

    if isinstance(dirs, list):
        log.warning("DeprecationWarning: `directives` are specified as list. We advice to use dictionaries instead.")
        for idir in dirs:
            for itag, ival in idir.items():
                if itag == "config":
                    isubConfigTag, isubConfigPath = ival
                    if isubConfigTag in conf:
                        raise KeyError(f'sub config with name "{isubConfigTag}" already exists. Rename subconfig')
                    conf[isubConfigTag] = yaml.load(open(isubConfigPath), Loader=yaml.Loader)

    elif isinstance(dirs, dict):
        for key, file in dirs.items():
            conf[key] = yaml.load(open(confPath.parent / file), Loader=yaml.Loader)

    return conf


def _correct_deprecated_fields(content, log):

    backwards_compatibility = {
        "actionFolder": "action_folder",
        "timeit": "time_it",
        "enable development mode": "enable_development_mode",
        "fromCheckPoint": "from_check_point",
    }

    refactorings = []
    for old, new in backwards_compatibility.items():
        if old in content:
            refactorings.append((old, new))
            content = content.replace(old, new)

    if len(refactorings) > 0:
        warn_msg = ", ".join((f"{i} => {j}" for i, j in refactorings))
        log.warning(
            f"deprecated fields in user.yaml detected. Please replace the following field names as follows: {warn_msg}"
        )

    return content


def validate_config(conf, confPath):
    """
    Raises error when user config violates specified schema. First a 'general_schema' is checked, which assures
    that user.yaml header is set correctly and each action in the workflow is defined with a 'project' and 'call'.
    Then, specified validations for 'call' == 'checkpoint' and all other actions are performed. If any of the
    validations is not successful, the procedure crashes logging the Exception
    """

    v_general = cerberus.Validator(general_schema)
    v_action = cerberus.Validator(action_schema)
    v_checkpoint = cerberus.Validator(checkpoint_schema)
    v_executable = cerberus.Validator(executable_schema)

    checks = list()
    validation_violation = list()

    if v_general.validate(conf):
        checks.append(True)

        for item in conf["workflow"]:
            wrapped_item = dict()
            wrapped_item["action"] = item

            if list(item.values())[0]["call"] == "checkpoint":
                if v_checkpoint.validate(wrapped_item):
                    checks.append(True)
                else:
                    checks.append(False)
                    validation_violation.append(v_checkpoint.errors)
            elif "executable" in list(item.values())[0].keys():
                if v_executable.validate(wrapped_item):
                    checks.append(True)
                else:
                    checks.append(False)
                    validation_violation.append(v_executable.errors)
            else:
                if v_action.validate(wrapped_item):
                    checks.append(True)
                else:
                    checks.append(False)
                    validation_violation.append(v_action.errors)
    else:
        checks.append(False)
        validation_violation.append(v_general.errors)

    if len(validation_violation) > 0:
        raise ConfigurationError('in config file "{}":\n{}'.format(confPath, validation_violation))


class ConfigProvider:
    """
    The ConfigProvider triggers the reading and validation of the config file.
    For this purpose it sets the pathes, parses the config file for a user schema
    and triggers the validation process.
    """

    def __init__(self):
        self.config = None
        self.userConfigPath = None

    def setPathes(self, userConfigPath):
        self.userConfigPath = userConfigPath

    def parse(self, overridedata={}):
        self.config = ConfigDict()
        self.config["user_yaml_path"] = pt.Path(self.userConfigPath)
        self.config["user"] = loadAndValidate(self.userConfigPath, overridedata=overridedata)

        return self.config

    def get(self, overridedata={}):
        if self.config is None:
            self.parse(overridedata)
        return self.config


configProvider = ConfigProvider()



#!/usr/bin/env python
# -*- coding:utf-8 -*-
__author__ = ["Benjamin Fuchs", "Judith Vesper", "Felix Nitsch"]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = ["Niklas Wulff", "Hedda Gardian", "Gabriel Pivaro", "Kai von Krbek"]

__license__ = "MIT"
__version__ = "1.0.1"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


"""
This file features several default scripts for setting up and testing ioProc.
"""

defaultRunPyFile = """
#-*- coding:utf-8 -*-


import ioproc.runners as run

run.start()
"""

defaultSetupFoldersPyFile = """
#-*- coding:utf-8 -*-


import ioproc.runners as run

run.create_folders()
"""

defaultUserContent = """actionFolder: ../../actions

debug:
  time_it: False
  enable_development_mode: False

from_check_point: start

workflow:
    # configure your workflow here
"""

defaultActions = r"""
#-*- coding:utf-8 -*-

from enum import Enum

import pandas as pd

import pathlib as pt
import subprocess as spr
import tempfile

from ioproc.tools import action
from ioproc.logger import mainlogger


@action('general')
def execute(dmgr: dict, config: dict, params: dict):
    '''
    executes an executable program on this local computer.

    :param dmgr:
    :param config:
    :param params:
    :return: nothing
    '''

    args = [params['executable'], ]

    provided_args = list(params.get('args', []))

    args.extend(provided_args)

    shell = params.get('shell', False)
    cwd = params.get('cwd', pt.Path.cwd().as_posix())

    mainlogger.info('executing ', *args)
    with tempfile.TemporaryDirectory() as tmpdirname:
        tmpdirname = pt.Path(tmpdirname)
        std = tmpdirname / 'stdout.dat'
        err = tmpdirname / 'errout.dat'
        with std.open('w') as stdoutF:
            with err.open('w') as erroutF:
                try:
                    status = spr.run(args, stdout=stdoutF, stderr=erroutF,
                                     shell=shell, cwd=cwd,
                                     encoding='utf-8', text=True,
                                     )
                except FileNotFoundError as exc:
                    if not shell:
                        mainlogger.error('maybe the error is related to shell=False. Please read the subprocess '
                                         'package documentation on the "shell" parameter for the function "run"'
                                         'for more information.')
                    raise exc

        mainlogger.info(f"executed {params['args']['executable']}")
        if status.returncode == 0:
            mainlogger.info('successfully')

            with std.open('r') as ipf:
                data = ipf.read()
            mainlogger.info(data)
        else:
            with std.open('r') as ipf:
                data = ipf.read()
            mainlogger.info(data)
            with err.open('r') as ipf:
                data = ipf.read()
            mainlogger.error('with errors!')
            mainlogger.error(data)



@action('general')
def print_data(dmgr, config, params):
    '''
    simple debugging printing function. Prints all data in the data manager.

    Does not have any parameters.
    '''
    for k, v in dmgr.items():
        mainlogger.info(k+' = \n'+str(v))


@action('general')
def checkpoint(dmgr, config, params):
    '''
    creates a checkpoint file in the current working directory with name
    Cache_TAG while TAG is supplied by the action config.

    :param tag: the tag for this checkpoint, this can never be "start"
    '''
    assert params['tag'] != 'start', 'checkpoints can not be named start'
    dmgr.toCache(params['tag'])
    mainlogger.info('set checkpoint "{}"'.format(params['tag']))


@action('general')
def parse_excel(dmgr, config, params):
    '''
    Parses given `excelFile` for specified `excelSheets` as dataframe object and stores it in the datamanager by the
    key specified in `write_to_dmgr`.
    `excelHeader` can be set to `True` or `False`.

    The action may be specified in the user.yaml as follows:
        - action:
            project: general
            call: parse_excel
            data:
              read_from_dmgr: null
              write_to_dmgr: parsedData
            args:
              excelFile: spreadsheet.xlsx
              excelSheet: sheet1
              excelHeader: True
    '''

    args = params['args']
    file = get_field(args, 'excelFile')
    excel_sheet = get_excel_sheet(args)
    header = get_header(get_field(args, 'excelHeader'))
    parsed_excel = pd.read_excel(io=file, sheet_name=excel_sheet, header=header)

    with dmgr.overwrite:
        dmgr[params['data']['write_to_dmgr']] = parsed_excel


def get_field(dictionary, key):
    ''' Returns value from given `dictionary` and raises `KeyError` if `key` is not specified '''
    try:
        return dictionary[key]
    except KeyError:
        message = "Missing key '{}'. Given keys are '{}'.".format(key, [key for key in dictionary.keys()])
        raise KeyError(message)


def get_excel_sheet(params):
    ''' Returns a list of excel_sheets to parse from given dictionary or raises error when field is not specified '''
    try:
        return get_field(params, 'excelSheet')
    except KeyError:
        message = "Please specify the Excel sheet(s) to parse in a list using under the field `excelSheets`."
        raise IOError(message)


class Header(Enum):
    TRUE = 0
    FALSE = None


def get_header(string):
    ''' Returns converted operator from given bool '''
    header_map = {True: Header.TRUE.value,
                  False: Header.FALSE.value,
                  }
    return header_map[string]
"""

defaultConfigContent = """
userschema: schema/user_schema.yaml
"""



#!/usr/bin/env python
# -*- coding:utf-8 -*-
import os
import pathlib as pt

from ioproc.driver import setupworkflow, setupfolders, execute


__author__ = ["Benjamin Fuchs"]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = [
    "Felix Nitsch",
    "Judith Vesper",
    "Niklas Wulff",
    "Hedda Gardian",
    "Gabriel Pivaro",
    "Kai von Krbek",
]

__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


envPath = pt.Path()

pathvar = os.environ["PATH"]
elem = pathvar.split(";")

for ielem in elem:
    if "Scripts" in ielem:
        envPath = pt.Path(ielem).parent
        break


def create_folders(workflowName="yourTestWorkflow"):
    """
    Creates the required folder structure.
    """
    setupfolders(workflowName)


def create_workflow(path=pt.Path.cwd()):
    """
    Creates a new workflow in the current work directory.
    """
    setupworkflow(path)


def start(useryaml=None):
    """
    Executes the workflow manager.
    """
    execute(useryaml)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
import pathlib as pt
import shutil
from functools import wraps

import arrow as ar
import yaml

from ioproc.actionmanager import getActionManager
from ioproc.datamanager import DataManager
from ioproc.defaults import defaultActions, defaultRunPyFile, defaultSetupFoldersPyFile, defaultUserContent
from ioproc.logger import mainlogger as log
from ioproc.exceptions import ActionParamsError


__author__ = ["Benjamin Fuchs", "Judith Vesper", "Kai von Krbek"]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = [
    "Felix Nitsch",
    "Niklas Wulff",
    "Hedda Gardian",
    "Gabriel Pivaro",
    "Kai von Krbek",
]

__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


def setupFolderStructure(workflowName=""):
    """
    Set up folder structure, which is required for use of the ioProc workflow manager.
    If path already exists, a warning is raised.
    """
    if workflowName.strip() == "":
        workflowName = "workflow1"

    workflowName = workflowName.replace(" ", "_")

    CWD = pt.Path.cwd()

    path2actions = pt.Path(CWD, "actions", "general.py")
    path2workflow = pt.Path(CWD, "workflows", workflowName)
    path2userConfig = pt.Path(path2workflow, "user.yaml")

    path2actions.parent.mkdir(parents=True, exist_ok=True)
    path2workflow.mkdir(parents=True, exist_ok=True)

    if path2userConfig.exists():
        log.warning(f"UserConfig already exists at location {path2userConfig}")
    else:
        try:
            with path2userConfig.open("w") as opf:
                opf.write(defaultUserContent)

            with path2userConfig.open("r") as ipf:
                d = yaml.load(ipf, Loader=yaml.SafeLoader)

            with path2userConfig.open("w") as opf:
                d["actionFolder"] = path2actions.parent.as_posix()
                yaml.dump(d, opf)
        except Exception as e:
            shutil.rmtree(path2userConfig.as_posix())
            raise e

    if path2actions.exists():
        log.warning(f"General.py already exists at location {path2actions.parent}")
    else:
        with path2actions.open("w") as opf:
            opf.write(defaultActions)

    root = pt.Path().cwd()
    folderScript = root / "create_folder_structure.py"
    if not folderScript.exists():
        with folderScript.open("w") as opf:
            opf.write(defaultSetupFoldersPyFile)

    workflowStartScript = path2workflow / "run.py"
    if not workflowStartScript.exists():
        with workflowStartScript.open("w") as opf:
            opf.write(defaultRunPyFile)


class action:
    def __init__(self, project):
        """
        decorator to register actions in the ActionManager and validate input
        :param project: name of project
        """
        self.projectName = project

    def __call__(self, f):
        actionMgr = getActionManager()

        @wraps(f)
        def __actionwrapper__(dmgr, config, params):
            log.debug("calling wrapper")

            if type(dmgr) != DataManager:
                raise TypeError("no valid dataManager instance passed")

            start = ar.now()

            try:
                ret = f(dmgr, config, params)
            except Exception as e:
                log.exception(f'Error occured calling action "{self.projectName}-{f.__name__}"')
                raise e

            end = ar.now()
            log.info(f"duration '{self.projectName}.{f.__name__}': {end - start}")
            return ret

        actionMgr[self.projectName, f.__name__] = __actionwrapper__

        log.debug("wrapping complete for action: " + str(__actionwrapper__.__name__))
        return __actionwrapper__


class ActionArgs(dict):
    def __init__(self, *args, **kwargs):
        self.parents = kwargs.pop('parents', ())
        if len(self.parents) == 0:
            self.parents = ""
        else:
            self.parents = 'under section [' + "][".join(self.parents) + '] '
        
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):
        if key in self:
            return super().__getitem__(key)

        message = "Parameter '{}' is not a valid action input. Possible action input parameter/s {}is/are {}. Please check the user.yaml.".format(key, self.parents, list(self.keys()))
        raise ActionParamsError(message)
   
    def __setitem__(self, key, value):
        raise TypeError('Action arguments are read only.')


def freeze(d, parents=None):
    """
    This function converts a given dictionary in a read-only mapping, also called frozen-dictionary.
    :return a frozen dictionary where a key cannot be added nor removed
    """
    if parents is None:
        parents = tuple()

    for key, value in d.items():
        if not hasattr(value, "lstrip") and not hasattr(value, "items") and hasattr(value, "__iter__"):
            d[key] = tuple(value)
        elif hasattr(value, "items"):
            new_parents = list(parents)
            new_parents.append(key) 
            new_parents = tuple(new_parents)
            d[key] = freeze(value, new_parents)
    return ActionArgs(d, parents=parents)


def humanize_time(dt):
    # based on: https://raimasoftware.com/how-to-humanize-date-and-time-in-python
    # get the interval from the < dt > to the current moment in seconds
    interval = int(dt.total_seconds())

    # edge case of < dt> is now
    if interval == 0:
        return f"{round(dt.total_seconds(), 2)} ms"
    else:
        interval = -interval

    names = ["second", "minute", "hour", "day", "week", "month", "year"]
    multipliers = [
        1,
        60,
        3600,
        3600 * 24,
        3600 * 24 * 7,
        3660 * 24 * 30,
        3660 * 24 * 30 * 12,
    ]

    pretty = ""

    for i, m in enumerate(multipliers[::-1]):
        value = interval // m
        interval -= value * m
        if value == 0:
            pass
        elif value == 1:
            pretty += f"{value} {names[-(i+1)]} "
        else:
            pretty += f"{value} {names[-(i+1)]}s "
    return pretty



#!/usr/bin/env python
# -*- coding:utf-8 -*-
import importlib.util
import os
import pathlib as pt
import sys

from ioproc.config import configProvider

__author__ = ["Benjamin Fuchs", "Judith Vesper", "Felix Nitsch"]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = ["Niklas Wulff", "Hedda Gardian", "Gabriel Pivaro", "Kai von Krbek"]

__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


class ActionError(Exception):
    """
    Error raised by the ActionManager
    """

    pass


class ActionManager(dict):
    """
    Behaves like a dictionary and stores the project name and the action name together with a reference to the action.
    """

    def __setitem__(self, key, value):
        """
        overloaded method to handle two level keys.
        :param key: a tuple containing the project name and the action name
        :param value: the reference to an action
        """
        projectName, actionName = key

        if projectName not in self:
            super().__setitem__(projectName, {})

        if actionName in self[projectName]:
            raise ActionError("Action {} already registered.".format(actionName))

        self[projectName][actionName] = value


class __actioninit:
    """
    Asserts that only one ActionManager instance exists in the framework (Singleton).
    Within the framework, getActionManager is called to get a new ActionManager instance.
    getActionManager is an instance of the __actioninit class. On calling the getActionManager
    a check is done, if an ActionManager was already created. In this case the ActionManager
    instance is returned. If there is no instantiated ActionManager, a new instance is created
    and populated with actions.
    """

    actionMgr = None

    def __call__(self, config=None):
        """
        A call of the class __actioninit checks if an action manager instance is already available. (callable class)
        If no, it instantiates and populates a new one.
        If yes, it returns the previously created instance.
        :return: populated ActionManager instance
        """
        if self.actionMgr is None:
            self.actionMgr = ActionManager()
            if config is None:
                config = configProvider.get()

            p = os.path.abspath(
                config["user_yaml_path"].parent / config["user"]["action_folder"]
            )
            sys.path.append(p)
            action_folder_path = pt.Path(p)

            for ifile in action_folder_path.glob("./*.py"):
                spec = importlib.util.spec_from_file_location(
                    f"actions.{ifile.name.strip('.py')}", ifile.as_posix()
                )
                __m = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(__m)

            for imodule in action_folder_path.glob("./*/__init__.py"):
                module_name = imodule.parent.name
                for ifile in imodule.parent.glob("./**/*.py"):
                    spec = importlib.util.spec_from_file_location(
                        f"actions.{module_name}.{ifile.name.strip('.py')}",
                        ifile.as_posix(),
                    )
                    __m = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(__m)

        return self.actionMgr


getActionManager = __actioninit()



#!/usr/bin/env python
# -*- coding:utf-8 -*-

__author__ = ["Benjamin Fuchs", "Jan Buschmann", "Felix Nitsch"]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = []

__license__ = "MIT"
__version__ = "1.0.4"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


general_schema = {
    "workflow": {
        "required": True,
        "type": "list",
        "schema": {
            "type": "dict",
            "valuesrules": {
                "type": "dict",
                "schema": {
                    "project": {"type": "string", "required": True,},
                    "call": {"type": "string", "required": True,},
                    "data": {"required": False,},
                    "args": {"required": False,},
                    "tag": {"required": False,},
                    "executable": {"required": False,},
                    "shell": {"required": False,},
                    "cwd": {"required": False,},
                },
            },
        },
    },
    "action_folder": {"type": "string", "required": True,},
    "debug": {
        "type": "dict",
        "required": True,
        "schema": {
            "time_it": {"type": "boolean", "required": True,},
            "enable_development_mode": {"type": "boolean", "required": False,},
            "log_level": {
                "type": "string",
                "required": False,
                "allowed": ["INFO", "WARNING", "DEBUG", "CRITICAL", "ERROR"],
            },
        },
    },
    "from_check_point": {"type": ["string", "integer"],},
    "ignore_check_points": {"type": ["string", "integer"], "required": False,},
    "global": {"type": ["string", "integer", "float", "dict", "list"], "required": False,},
}


action_schema = {
    "action": {
        "type": "dict",
        "keysrules": {"type": "string",},
        "valuesrules": {
            "type": "dict",
            "schema": {
                "project": {"type": "string", "required": True,},
                "call": {"type": "string", "required": True,},
                "data": {
                    "required": False,
                    "schema": {
                        "read_from_dmgr": {
                            "nullable": True,
                            "type": ["string", "list", "dict"],
                            "required": True,
                            "forbidden": ["None", "none"],
                        },
                        "write_to_dmgr": {
                            "nullable": True,
                            "type": ["string", "list", "dict"],
                            "required": True,
                            "forbidden": ["None", "none"],
                        },
                    },
                },
                "args": {"type": "dict", "required": False,},
            },
        },
    }
}


checkpoint_schema = {
    "action": {
        "type": "dict",
        "keysrules": {"type": "string",},
        "valuesrules": {
            "type": "dict",
            "schema": {
                "project": {"type": "string", "required": True,},
                "call": {"type": "string", "required": True,},
                "tag": {"type": ["string", "integer"], "required": True,},
            },
        },
    }
}


executable_schema = {
    "action": {
        "type": "dict",
        "keysrules": {"type": "string",},
        "valuesrules": {
            "type": "dict",
            "schema": {
                "project": {"type": "string", "required": True,},
                "call": {"type": "string", "required": True,},
                "executable": {"required": True, "type": "string",},
                "shell": {"required": False, "type": "boolean",},
                "cwd": {"required": False, "type": "string",},
                "args": {"type": "dict", "required": False,},
            },
        },
    }
}



#!/usr/bin/env python
# -*- coding:utf-8 -*-

__author__ = ["Benjamin Fuchs"]
__copyright__ = "Copyright 2021, German Aerospace Center (DLR)"
__credits__ = []

__license__ = "MIT"
__version__ = "1.0.2"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


class UnknownActionModule(Exception):
    pass


class UnknownAction(Exception):
    pass


class CheckPointError(Exception):
    pass


class ActionParamsError(Exception):
    pass


class MissingOverridesError(Exception):
    pass



#!/usr/bin/env python
# -*- coding:utf-8 -*-
import os
import pathlib as pt

import arrow as ar
import click

from ioproc.actionmanager import getActionManager
from ioproc.config import configProvider
from ioproc.datamanager import DataManager
from ioproc.defaults import defaultConfigContent, defaultRunPyFile, defaultUserContent
from ioproc.exceptions import CheckPointError, UnknownAction, UnknownActionModule
from ioproc.logger import datalogger as dlog
from ioproc.logger import mainlogger as log
from ioproc.tools import freeze, humanize_time, setupFolderStructure

__author__ = ["Benjamin Fuchs", "Judith Vesper", "Felix Nitsch"]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = ["Niklas Wulff", "Hedda Gardian", "Gabriel Pivaro", "Kai von Krbek"]

__license__ = "MIT"
__version__ = "1.0.2"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


IOPROCINSTALLROOT = pt.Path(__file__).resolve().parent
SCHEMAPATH = pt.Path(IOPROCINSTALLROOT, "schema")
HOME = pt.Path.home()

defaultConfigContent = defaultConfigContent.format(IOPROCINSTALLROOT.as_posix())


@click.group()
@click.pass_context
def ioproc(ctx):
    ctx.ensure_object(dict)


@ioproc.command(help="setup default folder structure")
@click.option("--workflowname", default="workflow1", help="name of your workflow")
def setupfolders(workflowname):
    setupFolderStructure(workflowname)


@ioproc.command(help="create all necessary files for workflow in current folder or at userconfig location")
@click.option("--path", "-p", default=None, help="path to folder where to create workflow")
def setupworkflow(path):
    path = pt.Path.cwd() if path is None else pt.Path(path)
    if not path.exists():
        raise IOError(f"workflow folder not found: {path.as_posix()}")

    p = path / "user.yaml"
    if not p.exists():
        with p.open("w") as opf:
            opf.write(defaultUserContent)

    p = path / "run.py"
    if not p.exists():
        with p.open("w") as opf:
            opf.write(defaultRunPyFile)


def format_override(ctx, self, s):
    try:
        d = tuple(i.strip().split("=") for i in s)
        d = dict(d)
    except Exception:
        raise click.exceptions.BadArgumentUsage("the overwrites need to be of shape: ioproc overwrites a=1 b=2")

    return d


@ioproc.command(help="execute ioproc workflow defined by user.yaml in current folder")
@click.option("--useryaml", "-u", default=None, help="path to user.yaml")
@click.option(
    "--override/--no-override", "-o", default=False, help="override values of the user.yaml (based on jinja2 syntax)",
)
@click.argument("overridedata", nargs=-1, callback=format_override)
def execute(useryaml, override, overridedata):

    if override and not overridedata:
        raise click.exceptions.ClickException("Missing override data")
    elif not override and overridedata:
        raise click.exceptions.ClickException("overrides need to be specified with ioproc execute -o")

    _execute(useryaml, overridedata=overridedata, return_data=False)


def _execute(path_to_user_yaml: pt.Path, overridedata: [dict, None] = None, return_data: bool = False,) -> [dict, None]:
    start = ar.now()

    if overridedata is None:
        overridedata = {}

    userConfigPath = pt.Path(pt.Path.cwd(), "user.yaml")

    if path_to_user_yaml is not None:
        userConfigPath = pt.Path(path_to_user_yaml)

    old_pwd = pt.Path.cwd()

    os.chdir(userConfigPath.parent.as_posix())
    userConfigPath = userConfigPath.relative_to(userConfigPath.parent)

    try:
        data = _execute_workflow(userConfigPath, overridedata, return_data)
    finally:
        os.chdir(old_pwd.as_posix())
        end = ar.now()
        log.info(f"run duration: {humanize_time(end-start)}")

    if return_data:
        return data


def _execute_workflow(userConfigPath, overridedata, return_data):

    configProvider.setPathes(userConfigPath=userConfigPath,)

    config = configProvider.get(overridedata)

    log, dlog = _setup_workflow_logger(config)

    actionMgr = getActionManager(config)
    assert (
        len(actionMgr) > 0
    ), "ActionManager is not defined. Ensure 'actionFolder' path in 'user.yaml' is set correctly."
    dmgr = DataManager(config["user"]["debug"]["enable_development_mode"])

    log.info("starting workflow")

    log.debug("commencing action calling")

    FROM_CHECK_POINT = config["user"]["from_check_point"] != "start"
    IGNORE_CHECK_POINTS = False
    if "ignore_check_points" in config["user"]:
        IGNORE_CHECK_POINTS = config["user"]["ignore_check_points"]

    if IGNORE_CHECK_POINTS:
        FROM_CHECK_POINT = False

    _validate_workflow_actions(config, actionMgr)

    CHECK_POINT_DEFINED_IN_WORKFLOW = False

    if FROM_CHECK_POINT:
        _check_for_checkpoint_data(config)

    for iActionInfo in config["user"]["workflow"]:
        iActionInfo = iActionInfo[list(iActionInfo.keys())[0]]
        if FROM_CHECK_POINT:
            if "tag" in iActionInfo and iActionInfo["tag"] == config["user"]["from_check_point"]:
                FROM_CHECK_POINT = False
                CHECK_POINT_DEFINED_IN_WORKFLOW = True
                dmgr.fromCache(config["user"]["from_check_point"], iActionInfo)
                log.info('reading from cache for tag "{}"'.format(config["user"]["from_check_point"]))
            continue
        log.debug('executing action "' + iActionInfo["call"] + '"')
        dmgr.entersAction(iActionInfo)

        if IGNORE_CHECK_POINTS and iActionInfo["call"] == "checkpoint":
            log.warning(f'skipping checkpoint "{iActionInfo["tag"]}" due to "ignore check points" option set to true')
            continue

        try:
            actionMgr[iActionInfo["project"]][iActionInfo["call"]](dmgr, config, freeze(iActionInfo))
        except Exception as e:
            log.exception(
                'Fatal error during execution of action "'
                + iActionInfo["call"]
                + '":\nData manager log:\n'
                + dmgr.report()
            )
            raise e
        dmgr.leavesAction()

    if not CHECK_POINT_DEFINED_IN_WORKFLOW and FROM_CHECK_POINT:
        raise CheckPointError(
            f'The requested checkpoint "{config["user"]["from_check_point"]}" ' f"is not defined in the workflow!"
        )

    if return_data:
        return dmgr.export_to_dict()


def _check_for_checkpoint_data(config):
    given_tag = config["user"]["from_check_point"]
    if not pt.Path("./.checkpoint_data/Cache_{}.h5f".format(given_tag)).is_file():
        message = (
            f"cannot find cache file with given tag name '{given_tag}' "
            f" in './.checkpoint_data/Cache_{given_tag}.h5f'"
        )
        log.exception(message)
        raise Exception(message)


def _setup_workflow_logger(config):
    if "debug" in config["user"] and "log_level" in config["user"]["debug"]:
        log.setLevel(config["user"]["debug"]["log_level"])
        dlog.setLevel(config["user"]["debug"]["log_level"])
    else:
        log.setLevel("INFO")
        dlog.setLevel("INFO")
    return log, dlog


def _validate_workflow_actions(config, actionMgr):
    # this is a cross check loop to identify early problems in the workflow.
    for iActionInfo in config["user"]["workflow"]:
        iActionInfo = iActionInfo[list(iActionInfo.keys())[0]]
        if iActionInfo["project"] not in actionMgr:
            raise UnknownActionModule(
                f'The action module "{iActionInfo["project"]}" is not known to ioproc.\n'
                f'please check in the action folder for it: "{config["user"]["action_folder"]}"'
            )

        if iActionInfo["call"] not in actionMgr[iActionInfo["project"]]:
            raise UnknownAction(
                f'The action "{iActionInfo["call"]}" is not known to ioproc.\n'
                f'please check in the action folder "{config["user"]["action_folder"]}"\n'
                f'in the module {iActionInfo["project"]} for the action.'
            )


if __name__ == "__main__":
    ioproc()



#!/usr/bin/env python
# -*- coding:utf-8 -*-

import ioproc.driver


__author__ = [
    "Benjamin Fuchs",
]
__copyright__ = "Copyright 2020, German Aerospace Center (DLR)"
__credits__ = [
    "Judith Vesper",
    "Felix Nitsch",
    "Niklas Wulff",
    "Hedda Gardian",
    "Gabriel Pivaro",
    "Kai von Krbek",
]

__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Felix Nitsch"
__email__ = "ioProc@dlr.de"
__status__ = "Production"


if __name__ == "__main__":
    ioproc.driver.execute()
